// ignore_for_file: implementation_imports
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:analyzer/src/dart/element/element.dart';
import 'package:collection/collection.dart';
import 'package:flutterx_reflection_codegen/src/encoder/source_encoder.dart';
import 'package:flutterx_reflection_codegen/src/encoder/target_sdk.dart';
import 'package:flutterx_reflection_codegen/src/encoder/top_level/encoder_library.dart';

class CodegenConfig {
  static final Map<String, String> _defaultValues = {};
  static final Set<String> _generateFields = {};

  const CodegenConfig._();

  static void load(LibraryElement sdkLibrary) {
    final configurationVariable = sdkLibrary.topLevelElements
        .whereType<TopLevelVariableElement>()
        .firstWhereOrNull((element) => element.name == 'configuration');
    if (configurationVariable == null || !configurationVariable.isConst)
      return print(
          'Configuration variable not found. To provide a custom configuration add \'const configuration = { ... };\'');
    final configuration = _getMapNodes((configurationVariable as ConstVariableElement).constantInitializer);
    final defaultValues = configuration['defaultValues'];
    if (defaultValues != null) {
      final alias = LibraryEncoder.libraryAliasOf(sdkLibrary);
      for (final defaultValue in _getMapNodes(defaultValues).entries)
        CodegenConfig._defaultValues[defaultValue.key] = _computeDefaultValue(alias, defaultValue.value);
    }
    final generateFields = configuration['generateFields'];
    if (generateFields != null) {
      for (final referenceId in _getSetNodes(generateFields))
        CodegenConfig._generateFields.add((referenceId as SimpleStringLiteral).value);
    }
  }

  static String? defaultValueFor(ParameterElement parameter) => _defaultValues[parameterReferenceId(parameter)];

  static bool generateFieldsFor(ClassElement element) => _generateFields.contains(referenceId(element));

  static String _computeDefaultValue(String alias, Expression value) {
    if (value is Identifier) {
      final name = value.name;
      if (Identifier.isPrivateName(name))
        throw ArgumentError('Bad configuration syntax: default value identifier cannot be private, found $name');
      if (_isDefaultValueResolver(value.staticType))
        throw ArgumentError(
            'Bad configuration syntax: DefaultValueResolver must be declared directly inside \'defaultValues\' map, found $value');
      return '$alias.$name';
    }
    if (value is InstanceCreationExpression) {
      if (!_isDefaultValueResolver(value.constructorName.type.type))
        throw ArgumentError(
            'Bad configuration syntax: default values can only be identifiers or DefaultValueResolver, found $value');
      final arguments = value.argumentList.arguments;
      if (arguments.length != 1)
        throw ArgumentError(
            'Bad configuration syntax: DefaultValueResolver accepts 1 argument, found [${arguments.join(', ')}]');
      final identifier = arguments[0];
      if (identifier is! Identifier)
        throw ArgumentError(
            'Bad configuration syntax: DefaultValueResolver argument should be an identifier, found $identifier');
      return '$kDefaultValueResolver($alias.${identifier.name})';
    }
    throw ArgumentError(
        'Bad configuration syntax: default values can only be identifiers or DefaultValueResolver, found $value');
  }

  static Map<String, Expression> _getMapNodes(Expression? node) {
    if (node is! SetOrMapLiteral) throw ArgumentError('Bad configuration syntax: expected map literal, found $node');
    final entries = node.elements.whereType<MapLiteralEntry>();
    return Map.fromEntries(entries.map((entry) {
      final key = entry.key;
      if (key is! SimpleStringLiteral) throw ArgumentError('Bad configuration syntax: keys must be string literals');
      return MapEntry(key.value, entry.value);
    }));
  }

  static Set<CollectionElement> _getSetNodes(Expression? node) {
    if (node is! SetOrMapLiteral) throw ArgumentError('Bad configuration syntax: expected map literal, found $node');
    return node.elements.toSet();
  }

  static bool _isDefaultValueResolver(DartType? type) {
    final element = type?.element;
    return element is ClassElement && element.name == 'DefaultValueResolver';
  }
}
